import 'dart:io';

import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/generators/integration_generator.dart';
import 'package:welltested/generators/integration_saver.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/file_extensions.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/utilities/logging.dart';

import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/utilities.dart';

import '../src/models/generate_run_info.dart';
import '../src/models/test_request/integration/integration_test_input.dart';
import '../src/utilities/telemetry_client.dart';

class E2EOperation {
  Future<void> saveE2E({
    required PubspecData pubspecData,
    required String className,
    int? index,
  }) async {
    try {
      FileSystemEntity file = File(PathUtils.flowsPath);
      if (!file.existsSync()) {
        wtLog.error(
          'Welltested.dart file not found at ${file.path}.\n Please read our docs to learn more.',
        );
        return;
      }
      AnalysisContextCollection collection = AnalysisContextCollection(
        includedPaths: [
          PathUtils.integrationTestPath,
          File(PathUtils.flowsPath).absolute.path,
        ],
        resourceProvider: PhysicalResourceProvider.INSTANCE,
      );
      final dartServerRepository = DartServerRepository();
      await dartServerRepository.initialize();

      final libDirectory = Directory(PathUtils.libPath);
      final testDirectory = Directory(PathUtils.testPath);
      await dartServerRepository.addFilesToRoot(
          [libDirectory.absolute.path, testDirectory.absolute.path]);

      await processSaveIntegrationRequest(
        file,
        collection,
        pubspecData,
        dartServerRepository: dartServerRepository,
        className: className,
        index: index,
      );
      dartServerRepository.dispose();
    } catch (e, stackTrace) {
      //TODO: Dart Server Repository should still be disposed. (including in unit, widget, integration)
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning("Error saving integration tests");
      wtLog.log(e.toString());
    }
  }

  Future<void> generateE2E({
    required PubspecData pubspecData,
    required String className,
    int? index,
  }) async {
    try {
      final dartServerRepository = DartServerRepository();
      await dartServerRepository.initialize();
      final libDirectory = Directory(PathUtils.libPath);
      final integrationTestDirectory = Directory(PathUtils.integrationTestPath);
      final testRunInfo = GenerateRunInfo.newRun();

      await dartServerRepository.addFilesToRoot(
          [libDirectory.absolute.path, integrationTestDirectory.absolute.path]);

      FileSystemEntity file = File(PathUtils.flowsPath);
      if (!file.existsSync()) {
        wtLog.error(
          'Welltested.dart file not found at ${file.path}.\n Please read our docs to learn more.',
        );
        return;
      }
      AnalysisContextCollection collection = AnalysisContextCollection(
        includedPaths: [
          PathUtils.integrationTestPath,
          File(PathUtils.flowsPath).absolute.path,
        ],
        resourceProvider: PhysicalResourceProvider.INSTANCE,
      );
      await processIntegrationRequest(file, collection, pubspecData,
          className: className,
          dartServerRepository: dartServerRepository,
          index: index,
          testRunInfo: testRunInfo);
      dartServerRepository.dispose();
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning("Error generating integration tests");
      wtLog.log(e.toString());
    }
  }

  Future<void> processIntegrationRequest(FileSystemEntity file,
      AnalysisContextCollection collection, PubspecData pubspecData,
      {required String className,
      required DartServerRepository dartServerRepository,
      int? index,
      required TestRunInfo testRunInfo}) async {
    wtLog.startSpinner("Getting ready...");
    try {
      final testInputs = await _prepareIntegrationTestInputs(file, collection);

      for (var test in testInputs) {
        if (test.moduleName != className) continue;

        for (var i = 0; i < test.testFlows.length; i++) {
          if (index != null && i != index) {
            continue;
          }

          final integrationGenerator =
              IntegrationGenerator(dartServerRepository: dartServerRepository)
                ..init();

          await integrationGenerator.generateForFlow(
              testInput: test,
              allTests: testInputs,
              index: i,
              pubspecData: pubspecData,
              collection: collection,
              testRunInfo: testRunInfo);
        }
      }

      if (testRunInfo.failed == 0 && testRunInfo.created > 0) {
        wtLog.stopSpinner(
            message:
                "✅ Succeeded. Please find the generated tests in `/integration_test` folder");
      }
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning(
          "Error 'generating' integration tests for file: ${file.path}");
      wtLog.log(e.toString());
    } finally {
      wtLog.stopSpinner();
      return;
    }
  }

  Future<void> processSaveIntegrationRequest(
    FileSystemEntity file,
    AnalysisContextCollection collection,
    PubspecData pubspecData, {
    required String className,
    required DartServerRepository dartServerRepository,
    int? index,
  }) async {
    wtLog.startSpinner("Getting ready...");
    try {
      final testInputs = await _prepareIntegrationTestInputs(file, collection);

      for (var test in testInputs) {
        if (test.moduleName != className) continue;

        for (var i = 0; i < test.testFlows.length; i++) {
          if (index != null && i != index) {
            continue;
          }
          final integrationSaver =
              IntegrationSaver(dartServerRepository: dartServerRepository)
                ..init();

          await integrationSaver.saveTestFlow(
              testInput: test,
              allTests: testInputs,
              index: i,
              pubspecData: pubspecData,
              collection: collection);
        }
      }
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning("Error saving integration tests for file: ${file.path}");
      wtLog.log(e.toString());
    } finally {
      wtLog.stopSpinner();
      return;
    }
  }

  Future<List<IntegrationTestInput>> _prepareIntegrationTestInputs(
      FileSystemEntity file, AnalysisContextCollection collection) async {
    final resolvedUnit = await file.resolveFile(collection);
    CompilationUnit unit = resolvedUnit.unit;

    List<IntegrationTestInput> testInputs = [];
    for (final declaration in unit.declarations.whereType<ClassDeclaration>()) {
      if (declaration.declaredElement?.supertype.toString() ==
          'WelltestedIntegrationTest') {
        final moduleName = declaration.name.toString();
        final setup = await declaration.declaredElement
            ?.setup(getResolvedAstNodeForElement);
        final testFlows = await declaration.declaredElement
            ?.testFlows(getResolvedAstNodeForElement);
        final testObjects = await declaration.declaredElement
            ?.testObjects(getResolvedAstNodeForElement);

        final integrationTest = IntegrationTestInput(
          setup: setup ?? '',
          testFlows: testFlows ?? [],
          testObjects: (testObjects ?? []),
          moduleName: moduleName,
        );

        testInputs.add(integrationTest);
      }
    }
    return testInputs;
  }
}
