import 'dart:async';
import 'dart:io';

import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/generators/unit_saver.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/file_extensions.dart';
import 'package:welltested/src/extensions/list_extension.dart';
import 'package:welltested/src/models/flag_generation_run_info.dart';
import 'package:welltested/src/models/generate_run_info.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/save_run_info.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/user_repository.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';
import 'package:welltested_annotation/welltested_annotation.dart';

import '../generators/unit_generator.dart';
import '../src/repositories/dart_server_repository.dart';
import '../src/utilities/utilities.dart';
import '../src/utilities/warnings.dart';
import '../src/visitors/class_visitor.dart';

class UnitOperation {
  int generationsCount = 0;

  static Future<void> saveUnit({required PubspecData pubspecData}) async {
    wtLog.startSpinner("Getting ready...");
    final userRepository = UserRepository();
    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();

    final libDirectory = Directory(PathUtils.libPath);
    final testDirectory = Directory(PathUtils.testPath);
    await dartServerRepository.addFilesToRoot(
        [libDirectory.absolute.path, testDirectory.absolute.path]);

    Iterable<FileSystemEntity> fileList = libDirectory
        .listSync(recursive: true)
        .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );

    List<Future> saveFutures = [];
    final testRunInfo = SaveRunInfo.newRun();
    for (var file in fileList) {
      saveFutures.add(
        runSaveUnit(
            file, collection, pubspecData, dartServerRepository, testRunInfo),
      );
    }
    wtLog.startSpinner("Saving Unit Tests");
    await Future.wait(saveFutures);
    String? sessionInfo;
    try {
      //Session Info Failure shouln't abrupt the user experience
      final sessionResult = await userRepository.getSessionInfo('save');
      sessionInfo = sessionResult.info;
    } catch (e) {}
    wtLog.stopSpinner(
        message: '✅ Succeeded. ${testRunInfo.updated} were saved.' +
            (sessionInfo != null ? '\n${sessionInfo}' : ''));
    dartServerRepository.dispose();
  }

  Future<void> generateUnit({
    required PubspecData pubspecData,
    required List<String> filePaths,
    required List<String> classNames,
    required List<String> methodNames,
    required FlagGenerationRunInfo flagGenerationRunInfo,
    bool dryrun = false,
    bool testcases = false,
    bool testability = false,
  }) async {
    wtLog.startSpinner("Getting ready...");
    final userRepository = UserRepository();
    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();

    final libDirectory = Directory(PathUtils.libPath);
    final testDirectory = Directory(PathUtils.testPath);
    await dartServerRepository.addFilesToRoot(
        [libDirectory.absolute.path, testDirectory.absolute.path]);

    Iterable<FileSystemEntity> fileList = filePaths.isNotEmpty
        ? filePaths.map((filePath) => File(filePath)).toList()
        : (libDirectory.listSync(recursive: true))
            .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );
    List<Future> generationFutures = [];
    final testRunInfo = GenerateRunInfo.newRun();
    for (var file in fileList) {
      generationFutures.add(
        runUnit(
            file, collection, pubspecData, dartServerRepository, testRunInfo,
            classNames: classNames,
            methodNames: methodNames,
            dryrun: dryrun,
            testcases: testcases,
            testability: testability,
            flagGenerationRunInfo: flagGenerationRunInfo),
      );
    }

    wtLog.startSpinner(dryrun
        ? 'Running in dryrun mode for ${methodNames.humanPreferredString(3)}'
        : testcases
            ? 'Generating testcases for ${methodNames.humanPreferredString(3)}'
            : testability
                ? 'Analyzing testability for ${methodNames.humanPreferredString(3)}'
                : 'Generating unit tests');

    await Future.wait(generationFutures);

    if (generationsCount == 0 ||
        flagGenerationRunInfo.classQueue.isNotEmpty ||
        flagGenerationRunInfo.functionQueue.isNotEmpty) {
      final classGenerationFailure = flagGenerationRunInfo.classQueue;
      final methodGenerationFailure = flagGenerationRunInfo.functionQueue;

      if (classGenerationFailure.isEmpty && methodGenerationFailure.isEmpty) {
        wtLog.warning("No class or method found with `@Welltested` annotation");
      } else {
        wtLog.warning(
            "The following classes/methods not found with `@Welltested` annotation:");
        if (classGenerationFailure.isNotEmpty)
          wtLog.warning(
              'Classes: ${classGenerationFailure.humanPreferredString(-1)}');
        if (methodGenerationFailure.isNotEmpty)
          wtLog.warning(
              'Methods: ${methodGenerationFailure.humanPreferredString(-1)}');
      }
    }

    if (!dryrun && !testcases && !testability) {
      String? sessionInfo;
      Map<String, dynamic>? cliVersionInfo;
      try {
        //Session Info Failure shouln't abrupt the user experience
        final sessionResult = await userRepository.getSessionInfo('generate');
        sessionInfo = sessionResult.info;
        cliVersionInfo = sessionResult.additionalInfo['cli_version'];
      } catch (_) {}

      if (cliVersionInfo != null && cliVersionInfo['is_critical'] == true) {
        wtLog.stopSpinner(
            message:
                "🚨 ${cliVersionInfo['info'] ?? 'Attention: Welltested CLI package is outdated! Please update your CLI.'}",
            severity: MessageSeverity.Error);
      } else {
        wtLog.stopSpinner(
            message:
                '✅ Succeeded. Please find the generated tests in `/test` folder.\nCreated: ${testRunInfo.created}\nUpdated:${testRunInfo.updated}\nFailed:${testRunInfo.failed}' +
                    (sessionInfo != null ? '\n\n${sessionInfo}' : ''));
        if (cliVersionInfo != null)
          wtLog.info(cliVersionInfo['info'] ??
              'Note: New version of CLI available for download.');
      }
    }

    if (wtLog.isBusy) {
      wtLog.stopSpinner();
    }

    dartServerRepository.dispose();
    if (!dryrun && !testcases && !testability) await runMockitoCommand();
  }

  Future<void> runUnit(
      FileSystemEntity file,
      AnalysisContextCollection collection,
      PubspecData pubspecData,
      DartServerRepository dartServerRepository,
      TestRunInfo testRunInfo,
      {required List<String> classNames,
      required List<String> methodNames,
      required bool dryrun,
      required bool testcases,
      required bool testability,
      required FlagGenerationRunInfo flagGenerationRunInfo}) async {
    try {
      final resolvedUnit = await file.resolveFile(collection);
      CompilationUnit unit = resolvedUnit.unit;

      final unitGenerator = UnitGenerator(
          dartServerRepository: dartServerRepository,
          cacheManager: CacheManager());
      unitGenerator.init();

      List<Future> generationFutures = [];
      for (final declaration
          in unit.declarations.whereType<ClassDeclaration>()) {
        final element = declaration.declaredElement!;
        //selection conditions
        if (!isAnnotationPresent(element)) continue;
        if (classNames.isNotEmpty) {
          final className = declaration.name.toString().toLowerCase();
          if (!classNames.contains(className)) continue;

          flagGenerationRunInfo.popFromQueue(FlagType.className, className);
        }

        if (element.checkIfAWidgetOrState)
          wtLog.warning(Warnings.annotationOnWidget(element.displayName));

        generationFutures.add(unitGenerator.generateForClass(
            classElement: element,
            annotation: element.metadata.firstWhere((element) =>
                element.element!.displayName ==
                Welltested().runtimeType.toString()),
            pubspecData: pubspecData,
            classAstNode: declaration,
            compilationUnit: unit,
            testRunInfo: testRunInfo,
            classFile: file as File,
            requestedFunctions: methodNames,
            dryrun: dryrun,
            testcases: testcases,
            testability: testability,
            flagGenerationRunInfo: flagGenerationRunInfo));
      }
      for (final declaration
          in unit.declarations.whereType<ExtensionDeclaration>()) {
        final element = declaration.declaredElement!;

        if (!isAnnotationPresent(element)) continue;
        if (classNames.isNotEmpty) {
          final className = declaration.name.toString().toLowerCase();
          if (!classNames.contains(className)) continue;

          flagGenerationRunInfo.popFromQueue(FlagType.className, className);
        }

        generationFutures.add(unitGenerator.generateForExtension(
            compilationUnit: unit,
            annotation: element.metadata.firstWhere((element) =>
                element.element!.displayName ==
                Welltested().runtimeType.toString()),
            extAstNode: declaration,
            extElement: element,
            pubspecData: pubspecData,
            extFile: file as File,
            testRunInfo: testRunInfo,
            requestedFunctions: methodNames,
            dryrun: dryrun,
            testcases: testcases,
            testability: testability,
            flagGenerationRunInfo: flagGenerationRunInfo));
      }
      for (final declaration
          in unit.declarations.whereType<FunctionDeclaration>()) {
        if (!isAnnotationPresent(declaration.declaredElement as Element))
          continue;

        if (methodNames.isNotEmpty) {
          final methodName = declaration.name.toString().toLowerCase();
          if (!methodNames.contains(methodName)) continue;

          flagGenerationRunInfo.popFromQueue(FlagType.methodName, methodName);
        }

        if (!(declaration.declaredElement != null &&
            declaration.declaredElement is FunctionElement)) {
          generationFutures
              .add(unitGenerator.generateForGlobalExeculatbleElement(
            methodElement: declaration.declaredElement!,
            methodAstNode: declaration,
            compilationUnit: unit,
            annotation: declaration.declaredElement!.metadata.firstWhere(
                (element) =>
                    element.element!.displayName ==
                    Welltested().runtimeType.toString()),
            pubspecData: pubspecData,
            methodFile: file as File,
            testRunInfo: testRunInfo,
            dryrun: dryrun,
            testcases: testcases,
            testability: testability,
          ));
          continue;
        }
        final element = declaration.declaredElement! as FunctionElement;
        generationFutures.add(unitGenerator.generateForGlobalMethod(
          methodElement: element,
          methodAstNode: declaration,
          compilationUnit: unit,
          annotation: element.metadata.firstWhere((element) =>
              element.element!.displayName ==
              Welltested().runtimeType.toString()),
          pubspecData: pubspecData,
          methodFile: file as File,
          testRunInfo: testRunInfo,
          dryrun: dryrun,
          testcases: testcases,
          testability: testability,
        ));
      }
      generationsCount += generationFutures.length;
      await Future.wait(generationFutures);
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.error('Unable to resolve file ${file.path}', verbose: true);
    }
  }

  static Future<void> runSaveUnit(
    FileSystemEntity file,
    AnalysisContextCollection collection,
    PubspecData pubspecData,
    DartServerRepository dartServerRepository,
    TestRunInfo testRunInfo,
  ) async {
    try {
      final resolvedUnit = await file.resolveFile(collection);
      CompilationUnit unit = resolvedUnit.unit;

      final saver = UnitSaver(
          dartServerRepository: dartServerRepository,
          cacheManager: CacheManager());
      saver.init();

      List<Future> saveFutures = [];

      for (final declaration
          in unit.declarations.whereType<ClassDeclaration>()) {
        final element = declaration.declaredElement!;

        if (!isAnnotationPresent(element)) continue;

        if (element.checkIfAWidgetOrState) continue;

        saveFutures.add(saver.saveForClass(
            classElement: element,
            annotation: element.metadata.firstWhere((element) =>
                element.element!.displayName ==
                Welltested().runtimeType.toString()),
            pubspecData: pubspecData,
            classAstNode: declaration,
            compilationUnit: unit,
            testRunInfo: testRunInfo,
            classFile: file as File));
      }
      for (final declaration
          in unit.declarations.whereType<ExtensionDeclaration>()) {
        final element = declaration.declaredElement!;

        if (!isAnnotationPresent(element)) continue;

        saveFutures.add(saver.saveForExtension(
            compilationUnit: unit,
            annotation: element.metadata.firstWhere((element) =>
                element.element!.displayName ==
                Welltested().runtimeType.toString()),
            extAstNode: declaration,
            extElement: element,
            pubspecData: pubspecData,
            extFile: file as File,
            testRunInfo: testRunInfo));
      }
      for (final declaration
          in unit.declarations.whereType<FunctionDeclaration>()) {
        if (!isAnnotationPresent(declaration.declaredElement as Element))
          continue;
        if (!(declaration.declaredElement != null &&
            declaration.declaredElement is FunctionElement)) {
          saveFutures.add(saver.saveForGlobalExecutableElement(
            methodElement: declaration.declaredElement!,
            methodAstNode: declaration,
            compilationUnit: unit,
            annotation: declaration.declaredElement!.metadata.firstWhere(
                (element) =>
                    element.element!.displayName ==
                    Welltested().runtimeType.toString()),
            pubspecData: pubspecData,
            methodFile: file as File,
            testRunInfo: testRunInfo,
          ));
          continue;
        }
        final element = declaration.declaredElement! as FunctionElement;
        saveFutures.add(saver.saveForGlobalMethod(
          methodElement: element,
          methodAstNode: declaration,
          compilationUnit: unit,
          annotation: element.metadata.firstWhere((element) =>
              element.element!.displayName ==
              Welltested().runtimeType.toString()),
          pubspecData: pubspecData,
          methodFile: file as File,
          testRunInfo: testRunInfo,
        ));
      }
      await Future.wait(saveFutures);
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.error('Unable to resolve file ${file.path}', verbose: true);
    }
  }
}

runMockitoCommand() async {
  wtLog.startSpinner("Generating mocks");
  await Future.delayed(Duration(seconds: 1));
  final process = await Process.run(
      "dart", ["run", "build_runner", "build", "--delete-conflicting-outputs"],
      runInShell: true);

  if (process.exitCode != 0) {
    wtLog.stopSpinner(
        message: "Failed to generate mocks:",
        severity: MessageSeverity.Warning);
    wtLog.log(process.stderr);
    wtLog.log(
        'Note: You may also directly use the command `dart run build_runner build` to generate the mocks after fixing the above errors.');
  } else {
    wtLog.stopSpinner(
        message: 'Successfully generated mocks.',
        severity: MessageSeverity.Info);
  }
}

Future<Map<String, MethodDeclaration>> generateForElement(
    Element element) async {
  final classVisitor = ClassVisitor();
  final Map<String, MethodDeclaration> methodAstNodes = {};

  element.visitChildren(classVisitor);
  for (final method in classVisitor.methods.entries) {
    final node =
        (await getResolvedAstNodeForElement(method.value)) as MethodDeclaration;
    methodAstNodes[method.key] = node;
  }
  return methodAstNodes;
}
