/*
 * Created on October 7th 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'dart:io';

import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/generators/widget_generator.dart';
import 'package:welltested/generators/widget_saver.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/file_extensions.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

import '../src/utilities/utilities.dart';
import 'unit_operation.dart';

class WidgetOperation {
  Future<void> saveWidget(PubspecData pubspecData) async {
    wtLog.startSpinner("Getting ready...");
    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();
    Iterable<FileSystemEntity> fileList = Directory(PathUtils.libPath)
        .listSync(recursive: true)
        .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );

    List<Future> saveFutures = [];
    final testRunInfo = TestRunInfo.newRun();

    for (FileSystemEntity file in fileList) {
      saveFutures.add(
        runSaveWidget(
          file,
          collection,
          pubspecData,
          dartServerRepository,
          testRunInfo,
        ),
      );
    }
    wtLog.startSpinner("Saving Widget Tests");
    await Future.wait(saveFutures);
    wtLog.stopSpinner(
        message:
            '✅ Succeeded. ${testRunInfo.updated} were saved.\nNote: Saving working tests helps improve future test generations. Keep it up!');
    dartServerRepository.dispose();
  }

  static Future<void> runSaveWidget(
    FileSystemEntity file,
    AnalysisContextCollection collection,
    PubspecData pubspecData,
    DartServerRepository dartServerRepository,
    TestRunInfo testRunInfo,
  ) async {
    try {
      final resolvedUnit = await file.resolveFile(collection);
      CompilationUnit unit = resolvedUnit.unit;
      final cacheManager = CacheManager();
      final saver = WidgetSaver(
          dartServerRepository: dartServerRepository,
          cacheManager: cacheManager);
      saver.init();

      final widgets = unit.declarations
          .whereType<ClassDeclaration>()
          .where((element) => element.declaredElement!.checkIfAWidgetOrState);

      for (final widget in widgets) {
        final element = widget.declaredElement!;
        final fileImport =
            'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(file.path)}';

        if (!isAnnotationPresent(element)) continue;
        await saver.saveTestsForWidget(
          widget: widget,
          widgetName: element.name,
          compilationUnit: unit,
          fileImport: fileImport,
          subjectFile: File(file.path),
          testRunInfo: testRunInfo,
        );
      }
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.error('Unable to resolve file ${file.path}', verbose: true);
    }
  }

  Future<void> generateWidget({String? filePath}) async {
    wtLog.info("Generating widget test");

    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();

    Iterable<FileSystemEntity> fileList = filePath != null
        ? [File(filePath)]
        : (Directory(PathUtils.libPath).listSync(recursive: true))
            .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );
    final testRunInfo = TestRunInfo.newRun();

    List<Future> generationFutures = [];
    for (FileSystemEntity file in fileList) {
      generationFutures.add(
        runWidget(
          file,
          collection,
          dartServerRepository,
          testRunInfo,
        ),
      );
    }

    wtLog.startSpinner("Generating Widget Tests");
    await Future.wait(generationFutures);
    wtLog.stopSpinner(
        message:
            '✅ Succeeded. Please find the generated tests in `/test` folder.\nCreated: ${testRunInfo.created}\nUpdated:${testRunInfo.updated}\nFailed:${testRunInfo.failed}');

    dartServerRepository.dispose();
    if (testRunInfo.created > 0 || testRunInfo.updated > 0)
      await runMockitoCommand();
  }

  Future<void> runWidget(
    FileSystemEntity file,
    AnalysisContextCollection collection,
    DartServerRepository dartServerRepository,
    TestRunInfo testRunInfo,
  ) async {
    try {
      final resolvedUnit = await file.resolveFile(collection);
      CompilationUnit unit = resolvedUnit.unit;

      final widgetGenerator = WidgetGenerator(
        dartServerRepository: dartServerRepository,
        cacheManager: CacheManager(),
      );
      widgetGenerator.init();
      for (var declaration in unit.declarations) {
        if (declaration is ClassDeclaration) {
          final element = declaration.declaredElement!;
          if (!element.checkIfAWidgetOrState) continue;
          if (!isAnnotationPresent(element)) continue;

          await widgetGenerator.generateForWidget(
              classDeclaration: declaration,
              compilationUnit: unit,
              subjectFile: file.path,
              testRunInfo: testRunInfo);
        }
      }
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.error('Unable to resolve file ${file.path}', verbose: true);
    }
  }
}
