/*
 * Created on November 15th 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */

import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/widget_test_executor.dart';
import 'package:welltested/src/extensions/file_extensions.dart';
import 'package:welltested/src/extensions/integration_extensions.dart';
import 'package:welltested/src/extensions/string_extensions.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_request/integration/integration_update_test_request.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/visitors/integration_test_visitor.dart';

import '../src/executors/integration_executor.dart';
import '../src/models/test_request/integration/integration_test_input.dart';
import '../src/utilities/telemetry_client.dart';

class IntegrationSaver {
  final DartServerRepository dartServerRepository;
  late final GenerationRepository generationRepository;

  ///Executors
  late final FixerExecutor fixerExecutor;
  late final WidgetTestExecutor widgetTestExecutor;
  void init() {
    generationRepository = GenerationRepository();
    fixerExecutor = FixerExecutor(dartServerRepository.server);
    widgetTestExecutor = WidgetTestExecutor();
  }

  IntegrationSaver({required this.dartServerRepository});

  Future<void> saveTestFlow(
      {required IntegrationTestInput testInput,
      required List<IntegrationTestInput> allTests,
      required int index,
      required PubspecData pubspecData,
      required AnalysisContextCollection collection}) async {
    try {
      final file =
          PathUtils.correspondingIntegrationTestFile(testInput.moduleName);
      final resolvedUnit = await file.resolveFile(collection);
      CompilationUnit unit = resolvedUnit.unit;
      String? fileContent = await file.getContent();
      if (fileContent == null) return;
      final IntegrationTestVisitor integrationTestVisitor =
          IntegrationTestVisitor(
              flow: testInput.testFlows[index], sourceCode: fileContent);
      unit.visitChildren(integrationTestVisitor);
      integrationTestVisitor.main?.functionExpression.body
          .visitChildren(integrationTestVisitor);
      final existingFlow = integrationTestVisitor.getFlow;
      if (existingFlow == null)
        return; //This particular test hasn't been saved yet
      var hasErrors =
          await fixerExecutor.hasCompilationErrors(file.absolute.path);
      if (hasErrors) {
        wtLog.log(
            "⚠️ File ${PathUtils.relatetiveLibFilePath(file.path)} has errors and will not be updated");
        return;
      }

      //remove all other flows
      List<String> codeToReplace = [];
      //remove all other flows
      for (final flow in integrationTestVisitor.allFlows) {
        if (flow.flowDescription !=
            integrationTestVisitor.getFlow?.flowDescription) {
          codeToReplace.add(fileContent.getPrettyCodeForNode(flow) + ';');
        }
      }
      codeToReplace.forEach((element) {
        fileContent = fileContent!.replaceAll(element, '');
      });

      //TODO: Remove Test Flows
      IntegrationUpdateRequestData integrationUpdateRequestData =
          await IntegrationExecutor().prepareUpdateRequest(
        testInput,
        allTests,
        index,
        pubspecData,
        integrationTestVisitor.helperFunctions,
      );
      wtLog.startSpinner("Updating Test for ${testInput.moduleName}");

      await generationRepository.updateIntegrationTests(
          integrationUpdateRequestData, fileContent!);

      wtLog.stopSpinner(message: '✅ Succeeded. The test was saved.');
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning(
          "Error saving integration tests module: ${testInput.moduleName}.");
      wtLog.log(e.toString());
    } finally {
      wtLog.stopSpinner();
    }
  }
}
