/*
 * Created on May 11th 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */

import 'dart:io';

import 'package:analyzer/dart/ast/ast.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/widget_test_executor.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

import '../src/utilities/logging.dart';

class WidgetSaver {
  final DartServerRepository dartServerRepository;
  final CacheManager cacheManager;

  WidgetSaver({required this.dartServerRepository, required this.cacheManager});
  late final GenerationRepository generationRepository;

  ///Executors
  late final FixerExecutor fixerExecutor;
  late final WidgetTestExecutor widgetTestExecutor;
  void init() {
    generationRepository = GenerationRepository();
    fixerExecutor = FixerExecutor(dartServerRepository.server);
    widgetTestExecutor = WidgetTestExecutor();
  }

  Future<void> saveTestsForWidget({
    required ClassDeclaration widget,
    required String widgetName,
    required CompilationUnit compilationUnit,
    required String fileImport,
    required File subjectFile,
    required TestRunInfo testRunInfo,
    ClassDeclaration? classAstNode,
    ExtensionDeclaration? extAstNode,
    String? className,
    Map<String, MethodDeclaration>? methodAstNodes,
  }) async {
    try {
      final testRequestData = await widgetTestExecutor.prepareRequest(
          classDeclaration: widget, compilationUnit: compilationUnit);

      final file = PathUtils.correspondingPermanentWidgetTestFile(
          widgetName, subjectFile.path);
      if (!(await file.exists())) return;
      final testIsSaved = await cacheManager.checkIfWidgetTestSaved(
        await testRequestData.codeInScrutity(excludeAnnotations: false),
        subjectFile.path,
        testRequestData.widgetName,
      );
      if (testIsSaved) return; //this test is already saved.

      var hasErrors =
          await fixerExecutor.hasCompilationErrors(file.absolute.path);
      if (hasErrors) {
        wtLog.log(
            "⚠️ File ${PathUtils.relatetiveLibFilePath(file.path)} has errors and will not be updated");
        return;
      }
      final fixedTest = await file.readAsString();
      await generationRepository.updateWidgetTest(testRequestData, fixedTest);
      await cacheManager.cacheWidgetTest(
          code: await testRequestData.codeInScrutity(excludeAnnotations: false),
          codeFilePath: subjectFile.path,
          test: fixedTest,
          testFilePath: file.path,
          widgetName: widgetName,
          saved: true);

      testRunInfo.incrementFromAction(TestGenerationStatus.update, widgetName);
    } catch (e, stackTrace) {
      wtTelemetry.trackError(
          severity: Severity.error, error: e, stackTrace: stackTrace);
      wtLog.warning("Error saving widget tests for widget: ${widgetName}");
      wtLog.log(e.toString());
    }
  }
}
