// ignore_for_file: unnecessary_cast

import 'dart:io';

import 'package:analyzer/dart/element/element.dart';
import 'package:path/path.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_request/integration/integration_test_request.dart';
import 'package:welltested/src/models/test_request/integration/integration_update_test_request.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/utilities.dart';

import '../models/test_request/integration/integration_test_input.dart';
import '../utilities/uuid.dart';

class IntegrationExecutor {
  Future<IntegrationRequestData> prepareRequest(
    IntegrationTestInput testInput,
    List<IntegrationTestInput> allTests,
    int index,
    PubspecData pubspecData,
  ) async {
    IntegrationRequestData integrationRequestData = IntegrationRequestData(
      activeScopes: await getActiveScope(allTests),
      moduleName: testInput.moduleName,
      precondition: testInput.setup,
      testFlow: testInput.testFlows[index],
      testObjects: testInput.testObjects,
      projectId: pubspecData.packageName,
    );
    return integrationRequestData;
  }

  Future<IntegrationUpdateRequestData> prepareUpdateRequest(
    IntegrationTestInput testInput,
    List<IntegrationTestInput> allTests,
    int index,
    PubspecData pubspecData,
    Map<String, List> helperFunctions,
  ) async {
    List<Map<String, String>> helperMethods = [];
    List<Map<String, String>> helperFiles = [];
    for (var entry in helperFunctions.entries) {
      final helperClassName = entry.key;
      final List methodsUsed = entry.value;
      for (var element in methodsUsed) {
        // get method type (global, static or class)
        String methodType;
        if (element is FunctionElement) {
          methodType = 'global';
        } else if ((element as MethodElement).isStatic) {
          methodType = 'static';
        } else {
          methodType = 'class';
        }

        final fileUri = element.source.uri;
        final file = File.fromUri(fileUri);
        final ast = await getResolvedAstNodeForElement(element);
        final offset = ast.offset;
        final end = ast.end;

        final file_code = file.readAsStringSync();
        final method_code = file_code.substring(offset, end);
        final file_path = relative(file.path, from: PathUtils.currentPath);
        helperMethods.add({
          "method_name": element.name,
          "type": methodType,
          "method": method_code,
          "method_code_hash": uniqueHex(method_code),
          "class": helperClassName,
          "file_path": file_path,
          "comment": element.documentationComment.toString(),
        });
        if (helperFiles.indexWhere((element) =>
                element['file_path'] == file_path &&
                element['class'] == helperClassName) ==
            -1) {
          helperFiles.add({
            "class": helperClassName,
            "file_path": file_path,
            "code": file_code,
            "code_hash": uniqueHex(file_code)
          });
        }
      }
    }

    IntegrationUpdateRequestData integrationRequestData =
        IntegrationUpdateRequestData(
            moduleName: testInput.moduleName,
            precondition: testInput.setup,
            testFlow: testInput.testFlows[index],
            testObjects: testInput.testObjects,
            projectId: pubspecData.packageName,
            helperMethods: helperMethods,
            helperFiles: helperFiles);
    return integrationRequestData;
  }

  Future<List<ActiveScope>> getActiveScope(
      List<IntegrationTestInput> tests) async {
    List<ActiveScope> activeScopes = [];

    for (IntegrationTestInput test in tests) {
      final moduleName = test.moduleName;
      final flows = test.testFlows;
      final List<String> activeHashses = [];
      for (var i = 0; i < flows.length; i++) {
        final flow = flows[i];
        final flowHash = uniqueHex(flow);
        activeHashses.add(flowHash);
      }
      ActiveScope activeScope = ActiveScope(
        moduleName: moduleName,
        flow_hashes: activeHashses,
        precondition: test.setup,
      );
      activeScopes.add(activeScope);
    }
    return activeScopes;
  }
}
