import 'dart:convert';

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:welltested/src/extensions/annotation_ext.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/string_extensions.dart';
import 'package:welltested/src/models/variable.dart';
import 'package:welltested/src/utilities/agent/extraction_agent.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/utilities.dart';
import 'package:welltested/src/visitors/stateful_widget_visitor.dart';

class WidgetTestRequest {
  final String widgetName;
  final String fileImportPath;
  final List<DartObject> supportingClasses;
  final ClassDeclaration classDeclaration;
  final CompilationUnit compilationUnit;
  final List<ClassDeclaration> fullWidgetDetail;
  List<String> extractedCode = [];
  bool forceGenerate;
  final Map<Variable, VariableDeclaration> variableDeclarations;
  final List<List<SimpleIdentifier>> callFlows;
  final List<ClassElement> classElements; //class elements used in the widget
  final List<EnumElement> enumElements; //enum elements used in the widget
  final List<MethodElement> staticMethods;
  final Map<String, MethodInvocation> methodInvocations;
  final List<SimpleIdentifier> simpleIdentifiers;
  final List<Element> typeArguments;
  WidgetTestRequest({
    required this.widgetName,
    required this.fileImportPath,
    required this.supportingClasses,
    required this.classDeclaration,
    required this.compilationUnit,
    required this.fullWidgetDetail,
    this.forceGenerate = false,
    required this.variableDeclarations,
    required this.callFlows,
    required this.classElements,
    required this.enumElements,
    required this.staticMethods,
    required this.methodInvocations,
    required this.simpleIdentifiers,
    required this.typeArguments,
  });

  List<String> get flowPointers {
    List<String> pointers = [];
    callFlows.forEach((flow) {
      if (flow.length > 1) {
        for (int i = 1; i < flow.length; i++) {
          final parameterField = flow[i].staticElement?.name;
          final parameterType =
              flow[i].staticType?.getDisplayString(withNullability: true);
          final parameterParent =
              flow[i - 1].staticType?.getDisplayString(withNullability: false);
          if ((flow[i - 1].staticType)!.isDartCoreList &&
              (flow[i - 1].staticType is InterfaceType)) {
            final parameterParent = (flow[i - 1].staticType as InterfaceType)
                .typeArguments
                .first
                .getDisplayString(withNullability: true);
            pointers
                .add("$parameterParent.$parameterField return $parameterType");
          } else {
            pointers
                .add("$parameterParent.$parameterField return $parameterType");
          }
        }
      } else {
        final parent = flow[0].parent.toString();
        final parameterType =
            flow[0].staticType?.getDisplayString(withNullability: true);
        pointers.add("$parent return $parameterType");
      }
    });
    return pointers;
  }

  Future<void> addExtractioncode(List<ExtractedCode> requirements) async {
    final classesToExtract = requirements
        .where((element) => element.operation == AgentOperation.getClass)
        .map((e) => e.name)
        .toList();
    final methodsToExtract = requirements
        .where((element) => element.operation == AgentOperation.getMethod)
        .map((e) => e.name)
        .toList();

    for (ClassElement classElement in classElements) {
      final name = classElement.name.toString();
      if (classesToExtract.contains(name)) {
        // if its a stateful widget, then we need to extract the state class as well
        if (classElement.checkIfAWidgetOrState) {
          final code = await _extractWidgetCode(classElement);
          extractedCode.add(code);
        } else {
          final element = await getResolvedAstNodeForElement(classElement);
          final code = getCleanCode(element as ClassDeclaration);

          extractedCode.add(code);
        }
        classesToExtract.remove(name);
      }
    }

    for (EnumElement enumElement in enumElements) {
      final name = enumElement.name;
      if (classesToExtract.contains(name)) {
        final element = await getResolvedAstNodeForElement(enumElement);
        extractedCode.add(element.toSource());
        classesToExtract.remove(name);
      }
    }

    for (MethodElement methodElement in staticMethods) {
      final name = methodElement.name.toString();
      if (methodsToExtract.contains(name)) {
        final element = await getResolvedAstNodeForElement(methodElement);
        extractedCode.add(element.toSource());
        methodsToExtract.remove(name);
      }
    }

    for (final entry in methodInvocations.entries) {
      final name = entry.key;
      if (methodsToExtract.contains(name)) {
        final element = entry.value.methodName.staticElement;
        if (element != null) {
          final resolvedAstNode = await getResolvedAstNodeForElement(element);
          extractedCode.add(resolvedAstNode.toSource());
          methodsToExtract.remove(name);
        }
      }
    }

    for (final identifier in simpleIdentifiers) {
      final element = identifier.staticElement;
      if (element != null) {
        if (classesToExtract.contains(element.name)) {
          final code = element.librarySource!.contents.data;
          extractedCode.add(code);
          classesToExtract.remove(element.name);
        } else if (methodsToExtract.contains(element.name)) {
          final code = element.librarySource!.contents.data;
          extractedCode.add(code);
          methodsToExtract.remove(element.name);
        }
      }
    }

    for (Element element in typeArguments) {
      if (classesToExtract.contains(element.name)) {
        final astNode = await getResolvedAstNodeForElement(element);
        extractedCode.add(astNode.toSource());
        classesToExtract.remove(element.name);
      }
    }
  }

  Future<String> _extractWidgetCode(ClassElement classElement) async {
    if (classElement.checkIfAWidgetOrState) {
      String code = '';
      StatefulWidgetFinder statefulWidgetFinder = StatefulWidgetFinder();
      final element = await getResolvedAstNodeForElement(classElement);
      element.visitChildren(statefulWidgetFinder);
      if (statefulWidgetFinder.classElements.isEmpty) {
        code = getCleanCode(element as ClassDeclaration);
      } else {
        code += getCleanCode(element as ClassDeclaration) + "\n";
        for (final classElement in statefulWidgetFinder.classElements) {
          final element = await getResolvedAstNodeForElement(classElement);
          code += getCleanCode(element as ClassDeclaration);
        }
      }
      return code;
    } else {
      throw Exception('Not a widget');
    }
  }

  List<String> get testcases {
    List<String> casesFromAnnotation = [];

    classDeclaration.metadata.testcasesAnnotations.forEach((e) {
      final testcasesArgument = e.arguments?.arguments.first as ListLiteral;
      if (testcasesArgument.elements.isNotEmpty) {
        casesFromAnnotation = testcasesArgument.elements
            .map((element) => (element as StringLiteral).stringValue ?? "")
            .toList();
      }
    });

    return casesFromAnnotation;
  }

  String codeInScrutity({bool excludeAnnotations = true}) {
    String code = "";

    compilationUnit.directives.forEach((directive) => directive
            is ImportDirective
        ? code = code + directive.toSource() + '\n' //add all import directives
        : null);
    code = code + "\n" + fullWidgetDetail.map((e) => e.toSource()).join('\n');
    if (excludeAnnotations) {
      /// remove all welltested package related annotation
      classDeclaration.metadata.packageAnnotations.forEach((element) {
        final annotationCode = element.toSource();
        code = code.replaceAll(annotationCode, '');
      });
    }
    final variables = variableDeclarations.keys.toList();
    for (var i = 0; i < variables.length; i++) {
      // before: final value = getValue()
      // after: final String value = getValue()
      final Variable variable = variables[i];
      if (!code.contains(variable.variableWithType)) {
        code = code.replaceFirst(variable.name, variable.variableWithType);
      }
    }

    return code;
  }

  List<String> supportingClassesCode() {
    List<String> codes = [];

    for (var sclass in supportingClasses) {
      // check if it is a class
      if (sclass.type!.element!.kind == ElementKind.CLASS) {
        final classElement = sclass.toTypeValue()!.element as ClassElement;
        // comaplete the code
        final code = classElement.source.contents.data;
        codes.add(code.shrink);
      }
    }
    return codes;
  }

  Map<String, dynamic> toJson() {
    return {
      "widgetName": widgetName,
      "fileImportPath": fileImportPath,
      "code": codeInScrutity(),
      'supportingClasses': extractedCode.map((e) => e).toList()
        ..addAll(supportingClassesCode()),
      "force_generate": jsonEncode(forceGenerate),
      "flow_pointers": jsonEncode(flowPointers),
      "testcases": jsonEncode(testcases),
    };
  }

  void printData() {
    wtLog.log('WidgetTestRequest:', verbose: true);
    wtLog.log('  widgetName: $widgetName', verbose: true);
    wtLog.log('  fileImportPath: $fileImportPath', verbose: true);
    wtLog.log('  forceGenerate: $forceGenerate', verbose: true);
    wtLog.log('  variableDeclarations: $variableDeclarations', verbose: true);
    wtLog.log('  callFlows: $callFlows', verbose: true);
    wtLog.log('  classElements: $classElements', verbose: true);
    wtLog.log('  enumElements: $enumElements', verbose: true);
    wtLog.log('  staticMethods: $staticMethods', verbose: true);
    wtLog.log('  methodInvocations: $methodInvocations', verbose: true);
    wtLog.log('  simpleIdentifiers: $simpleIdentifiers', verbose: true);
    wtLog.log('  typeArguments: $typeArguments', verbose: true);
    wtLog.log('\n', verbose: true);
    wtLog.log('Flow Pointers:', verbose: true);
    for (String pointer in flowPointers) {
      wtLog.log('  $pointer', verbose: true);
    }
    wtLog.log('\n', verbose: true);
    wtLog.log('Extractioncodes:', verbose: true);
    for (String code in extractedCode) {
      wtLog.log(code, verbose: true);
    }
    wtLog.log('\n', verbose: true);
    wtLog.log('Testcases:', verbose: true);
    for (String testcase in testcases) {
      wtLog.log('  $testcase', verbose: true);
    }
    wtLog.log('\n', verbose: true);
    wtLog.log('Code in Scrutity:', verbose: true);
    wtLog.log(codeInScrutity(), verbose: true);
  }
}
