import 'dart:io';
import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:welltested/src/executors/writer_executor.dart';
import 'package:welltested/src/extensions/file_extensions.dart';
import 'package:welltested/src/extensions/integration_extensions.dart';
import 'package:welltested/src/extensions/string_extensions.dart';
import 'package:welltested/src/models/test_request/integration/helper_action.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/visitors/class_ast_visitor.dart';

class IntegrationAppendAgent {
  WriterExecutor _writerExecutor;
  IntegrationAppendAgent(this._writerExecutor);
  Future<void> appendMethod(
    HelperAction action,
  ) async {
    try {
      final absoluteFilePath =
          PathUtils.absouteProjectFilePath(action.filePath);
      final file = File(absoluteFilePath);
      if (!(await file.exists())) {
        await file.create(recursive: true);
      }
      AnalysisContextCollection collection = AnalysisContextCollection(
        includedPaths: [
          PathUtils.integrationTestPath,
          File(PathUtils.flowsPath).absolute.path,
        ],
        resourceProvider: PhysicalResourceProvider.INSTANCE,
      );
      final resolvedUnit = await file.resolveFile(collection);
      CompilationUnit unit = resolvedUnit.unit;
      final fileContent = await file.readAsString();

      if (action.className == null) {
        // global method

        final functions = unit.declarations.where((element) =>
            element is FunctionDeclaration &&
            element.name.toString() == action.methodName);
        final content = action.method;
        if (functions.isEmpty) {
          final updatedContent = fileContent + "\n" + content;
          await _writerExecutor.processHelperFunctionGeneration(
              file.absolute.path, updatedContent);
        } else {
          final function = functions.first;
          final offset = function.offset;
          final end = function.end;
          final updatedContent = fileContent.replaceRange(offset, end, content);
          await _writerExecutor.processHelperFunctionGeneration(
              file.absolute.path, updatedContent);
        }
      } else {
        // class method

        wtLog.log(
            "Existing classes: ${unit.declarations.where((element) => element is ClassDeclaration).map((e) => (e as ClassDeclaration).name.toString())}",
            verbose: true);

        final helperClasses = unit.declarations
            .where((element) => (element is ClassDeclaration &&
                element.name.toString() == action.className))
            .toList();

        if (helperClasses.isNotEmpty) {
          // class exists

          final ClassDeclaration helperClass =
              helperClasses.first as ClassDeclaration;
          final classAstVistor = ClassAstVistor();
          helperClass.visitChildren(classAstVistor);
          wtLog.log(
              "Existing methods: ${classAstVistor.methodDeclarations.map((e) => e.name.toString())}",
              verbose: true);
          final declarations = classAstVistor.methodDeclarations
              .where((element) => element.name.toString() == action.methodName);
          if (declarations.isEmpty) {
            // method doesn't exists => append
            wtLog.log("Method ${action.methodName} doesn't exist, creating",
                verbose: true);

            final lastMethodDeclaration =
                classAstVistor.methodDeclarations.last;
            final endOffset = lastMethodDeclaration.end;
            // TODO: sepearate imports and methodcode

            final newMethodCode = action.method;

            final updatedContent = fileContent.replaceRange(
                endOffset, endOffset, "\n" + newMethodCode + "\n");

            await _writerExecutor.processHelperFunctionGeneration(
                file.absolute.path, updatedContent);
          } else {
            // TODO: Point of discussion - Ideallly we should ask the devs if
            // they want to replace the file  or retains older one
            // method already exists => Update
            wtLog.log("Method ${action.methodName} exists already=> Updating",
                verbose: true);

            final hasComments = declarations.first.beginToken.precedingComments;

            var offset = declarations.first.offset;
            if (hasComments != null) {
              offset = hasComments.offset;
            }
            final end = declarations.first.end;
            final updatedContent =
                fileContent.replaceRange(offset, end, action.method);
            await _writerExecutor.processHelperFunctionGeneration(
                file.absolute.path, updatedContent);
          }
        } else {
          // class doesn't exists
          wtLog.log("Class ${action.className} doesnt exist, creating new}",
              verbose: true);
          final content =
              createNewClassWithMethod(action.className!, action.method);

          final updatedContent = fileContent + "\n" + content;
          await _writerExecutor.processHelperFunctionGeneration(
              file.absolute.path, updatedContent);
        }
      }
    } catch (e, stackTrace) {
      throw Exception(
          "Failed adding helper method. Please try again or contact Welltested team. Please reach out to team@welltested.ai for resolution. Details: ${e} ${stackTrace}");
    }
  }

  String createNewClassWithMethod(String className, String methodCode) {
    return 'class $className {\n  $methodCode\n}';
  }

  Future<String> appendNewTest(
      {required File existingTestFile,
      required String incomingTest,

      ///@Only required when helper methods have to be extracted
      required String currentFlow}) async {
    if (!(await existingTestFile.exists()))
      await _writerExecutor.outputGeneration(
        existingTestFile.path,
        incomingTest,
      );
    String testContent = await existingTestFile.readAsString();
    var newVisitor =
        incomingTest.executedVisitor(existingTestFile.path, currentFlow);

    if (newVisitor.allFlows.isNotEmpty) {
      final incomingTestFlow = newVisitor.allFlows.first;
      final flowDescriptionNode = incomingTestFlow.argumentList.arguments.first;
      incomingTest = incomingTest.replaceNodeCode(
          flowDescriptionNode, '\"${currentFlow}\"');
      newVisitor =
          incomingTest.executedVisitor(existingTestFile.path, currentFlow);
    }

    if (newVisitor.allFlows.isNotEmpty) {
      for (final flow in newVisitor.allFlows) {
        final exitingVisitor =
            testContent.executedVisitor(existingTestFile.path, currentFlow);
        if (exitingVisitor.main == null)
          throw ("Couldn't work without existing main"); //TODO: Could be handled
        final matchingFlow = exitingVisitor.matchingFlow(flow.flowDescription);

        if (matchingFlow != null) {
          final flowCode =
              incomingTest.getPrettyCodeForNode(flow, addSemiColon: false);
          testContent = testContent.replaceNodeCode(matchingFlow, flowCode);
        } else {
          final flowCode =
              incomingTest.getPrettyCodeForNode(flow, addSemiColon: true);
          int offset = exitingVisitor.allFlows.isNotEmpty
              ? exitingVisitor.allFlows.last.end + 1
              : exitingVisitor.main!.end - 1;
          testContent = testContent.insertAtOffset(offset, flowCode);
        }
      }
    }

    if (newVisitor.imports.isNotEmpty) {
      for (final import in newVisitor.imports) {
        final exitingVisitor =
            testContent.executedVisitor(existingTestFile.path, currentFlow);
        final importCode = incomingTest.getPrettyCodeForNode(import);
        if (exitingVisitor.imports.isEmpty) {
          testContent = importCode + '\n' + testContent;
        } else {
          if (!testContent.contains(import.toSource())) {
            testContent = testContent.insertAtOffset(
                exitingVisitor.imports.last.end, importCode);
          }
        }
      }
    }

    if (newVisitor.methodsDeclared.isNotEmpty) {
      for (final method in newVisitor.methodsDeclared) {
        final exitingVisitor =
            testContent.executedVisitor(existingTestFile.path, currentFlow);
        final matchingMethod = exitingVisitor.matchingMethods(method.name);
        final matchingMethodCode = incomingTest.getPrettyCodeForNode(method);
        if (matchingMethod != null) {
          testContent =
              testContent.replaceNodeCode(matchingMethod, matchingMethodCode);
        } else {
          int offset = exitingVisitor.allFlows.isNotEmpty
              ? exitingVisitor.allFlows.first.offset - 1
              : exitingVisitor.main!.end - 1;
          testContent = testContent.insertAtOffset(offset, matchingMethodCode);
        }
      }
    }

    await _writerExecutor.outputGeneration(
      existingTestFile.path,
      testContent,
    );
    return testContent;
  }
}
