import 'dart:async';
import 'dart:io';

import 'package:args/command_runner.dart';
import 'package:path/path.dart' as p;
import 'package:welltested/cli/integration_test_operation.dart';
import 'package:welltested/cli/unit_operation.dart';
import 'package:welltested/src/extensions/pubspec_data_extension.dart';
import 'package:welltested/src/models/dependency_info.dart';
import 'package:welltested/src/models/flag_generation_run_info.dart';
import 'package:welltested/src/utilities/env_parser.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/pubspec_parser.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';
import 'package:welltested/src/utilities/utilities.dart';

import '../../cli/widget_test_operations.dart';
import 'cache_manager.dart';

class UnitCommand extends Command {
  bool dryrun = false;
  bool testcases = false;
  bool testability = false;
  UnitCommand(String parent) {
    argParser.addFlag('verbose', abbr: 'v', callback: (value) {
      wtLog.verbose = value;
    });
    argParser.addFlag('dry-run', callback: (value) {
      this.dryrun = value;
    });
    argParser.addFlag('cases', callback: (value) {
      this.testcases = value;
    });
    argParser.addFlag('testability', callback: (value) {
      this.testability = value;
    });
    if (parent == 'generate') {
      argParser.addMultiOption('file',
          abbr: 'f',
          help: 'relative paths of the selected files',
          valueHelp: 'bloc/login_bloc.dart');
      argParser.addMultiOption("class",
          abbr: "c",
          help: 'names of the annotated classes',
          valueHelp: 'LoginBloc,FeedBloc');
      argParser.addMultiOption("method",
          abbr: 'm',
          help: "names of the methods under annotated classes",
          valueHelp: 'loginWithGoogle,loginWithEmailAndPassword');
    }
  }
  @override
  String get name => 'unit';

  @override
  String get description => 'Subcommand to select unit test';
  Future<void> run() async {
    if (verify_two_true([dryrun, testcases])) {
      wtLog.error(
          'No two of flags `dryrun` and `testcases` can be passed together');
      return;
    }
    if (this.dryrun || this.testcases || this.testability) {
      final methodNames = argResults?['method'] as List<String>?;
      if (methodNames == null || methodNames.isEmpty) {
        wtLog.error("Method name is required for dry run.");
        return;
      }
    }

    String currentDirectory = PathUtils.currentPath;
    final pubspecData = await PubspecParser.forPath(currentDirectory);
    setPackage(pubspecData.welltestedInfo?.dependencyVersion,
        pubspecData.welltestedInfo?.dev);
    await CacheManager().initialize();

    if (parent?.name == 'generate') {
      final flagGenerationRunInfo = FlagGenerationRunInfo();
      final filePaths = argResults?['file'] as List<String>? ?? [];
      final classNames = (argResults?['class'] as List<String>? ?? [])
          .map((className) => className.toLowerCase())
          .toList();
      final methodNames = (argResults?['method'] as List<String>? ?? [])
          .map((methodName) => methodName.toLowerCase())
          .toList();
      flagGenerationRunInfo.pushRequestInQueue(FlagType.className, classNames);
      flagGenerationRunInfo.pushRequestInQueue(
          FlagType.methodName, methodNames);

      // if file path is given, add the current directory to it
      final absoluteFilePath = <String>[];
      if (filePaths.isNotEmpty) {
        for (var filePath in filePaths) {
          filePath = p.join(PathUtils.currentPath, filePath);
          if (!File(filePath).existsSync()) {
            wtLog.warning("File does not exist: $filePath");
          }
          absoluteFilePath.add(filePath);
        }
        if (absoluteFilePath.isEmpty) return;
      }
      await UnitOperation().generateUnit(
          pubspecData: pubspecData,
          filePaths: absoluteFilePath,
          classNames: classNames,
          methodNames: methodNames,
          dryrun: dryrun,
          testcases: testcases,
          testability: testability,
          flagGenerationRunInfo: flagGenerationRunInfo);
      ;
    } else if (parent?.name == 'save') {
      await UnitOperation.saveUnit(pubspecData: pubspecData);
    }
    await CacheManager().closeAndSettle();
  }
}

class GenerateCommmand extends Command {
  @override
  String get name => 'generate';
  @override
  String get description => 'Generate tests for the annotated code';

  GenerateCommmand() {
    addSubcommand(UnitCommand(name));
    addSubcommand(WidgetCommand());
    addSubcommand(IntegrationCommand());
  }
}

class SaveCommand extends Command {
  @override
  String get name => 'save';

  @override
  String get description => 'Save updates made to existing tests';

  SaveCommand() {
    addSubcommand(UnitCommand(name));
    addSubcommand(WidgetCommand());
    addSubcommand(IntegrationCommand());
  }
}

class IntegrationCommand extends Command {
  IntegrationCommand() {
    argParser.addFlag('verbose', abbr: 'v', callback: (value) {
      wtLog.verbose = value;
    });
    argParser.addOption('file',
        abbr: 'f',
        help: 'relative path of the file',
        valueHelp: 'bloc/login_bloc.dart');
    argParser.addOption("class",
        abbr: "c", help: 'name of the WelltestedIntegrationTest config class', valueHelp: 'LoginBloc');
    argParser.addOption('index',
        abbr: "i", help: 'index of the testflow', valueHelp: '1');
  }

  @override
  String get name => 'integration';

  @override
  String get description => 'collect data as json objects';

  @override
  FutureOr? run() async {
    final className = argResults?['class'];
    if (className == null) {
      wtLog.error("Class name of the test is required.");
      return;
    }
    final pubspecData = await PubspecParser.forPath(PathUtils.currentPath);
    int? index;
    try {
      index = int.parse(argResults?['index']);
    } catch (e) {
      index = null;
    }
    if (parent?.name == "generate") {
      await E2EOperation().generateE2E(
          pubspecData: pubspecData,
          className: argResults?['class'],
          index: index);
    } else {
      await E2EOperation().saveE2E(
          pubspecData: pubspecData,
          className: argResults?['class'],
          index: index);
    }
  }
}

class InitCommand extends Command {
  @override
  String get name => 'init';

  @override
  String get description => 'Initialize welltested in your project';

  @override
  Future<void> run() async {
    try {
      final pubspecData = await PubspecParser.forPath(PathUtils.currentPath);
      await _addDependency(
          "welltested_annotation", pubspecData.welltestedAnnotationInfo,
          isDev: false, isFlutter: pubspecData.isFlutter);
      await _addDependency("build_runner", pubspecData.build_runnerInfo,
          isDev: true, isFlutter: pubspecData.isFlutter);
      await _addDependency("mockito", pubspecData.mockitoInfo,
          isDev: true, isFlutter: pubspecData.isFlutter);
      if (pubspecData.isFlutter) {
        await _addDependency("flutter_test", pubspecData.flutterTestInfo,
            isDev: true, isSDK: true, isFlutter: true);
      } else {
        await _addDependency("test", pubspecData.testInfo,
            isDev: true, isSDK: false, isFlutter: false);
      }
      await _addKeyToEnv();
      wtLog.info("Welltested initialized ✅");
    } catch (e) {
      wtLog.error("Error initializing welltested");
    }
  }

  _addDependency(String dependency, DependencyInfo? dependencyInfo,
      {bool isDev = false, bool isSDK = false, bool isFlutter = false}) async {
    final command = isFlutter ? "flutter" : "dart";

    if (dependencyInfo == null) {
      await runCommand(
          command,
          ["pub", "add", dependency]
            ..addAll(isSDK ? ["--sdk=flutter"] : [])
            ..addAll(isDev ? ["--dev"] : []));
    } else if (isDev && !dependencyInfo.dev) {
      await runCommand(command, ["pub", "remove", dependency]);
      await runCommand(
          command,
          ["pub", "add", dependency]
            ..addAll(isSDK ? ["--sdk=flutter"] : [])
            ..add(isDev ? "--dev" : ""));
      ;
    }
  }

  _addKeyToEnv() async {
    final DotEnv env = DotEnv(includePlatformEnvironment: false)..load();
    final apiKey = env.getOrElse("WELLTESTED_API", () => "");
    if (apiKey.isEmpty) {
      wtLog.info("Please enter your welltested api key received in email");
      stdout.write(">> ");
      String? key = stdin.readLineSync();
      if (key == null || key.isEmpty) {
        wtLog.error("Api key can't be empty");
        return;
      }
      env.addNew({"WELLTESTED_API": key});
      wtLog.info("API Key successfully added.");
    }
  }
}

// widget test command
class WidgetCommand extends Command {
  @override
  String get description => 'Subcommand to select widget test';

  @override
  String get name => 'widget';

  WidgetCommand() {
    argParser.addFlag('verbose', abbr: 'v', callback: (value) {
      wtLog.verbose = value;
    });
    argParser.addOption('file',
        abbr: 'f',
        help: 'relative path of the file',
        valueHelp: 'screens/homescreen.dart');
  }

  @override
  Future<void> run() async {
    WidgetOperation widgetOperation = WidgetOperation();
    await CacheManager().initialize();

    if (parent?.name == 'generate') {
      await widgetOperation.generateWidget(
        filePath: argResults?['file'],
      );
    } else if (parent?.name == 'save') {
      String currentDirectory = PathUtils.currentPath;
      final pubspecData = await PubspecParser.forPath(currentDirectory);
      await widgetOperation.saveWidget(pubspecData);
    }
    await CacheManager().closeAndSettle();
  }
}
