import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/ast/visitor.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:welltested/src/extensions/integration_extensions.dart';
import 'package:welltested/src/utilities/path_utils.dart';

class IntegrationTestVisitor extends RecursiveAstVisitor {
  IntegrationTestVisitor({required this.flow, required String sourceCode})
      : _helperMethodVisitor = HelperMethodVisitor(sourceCode);
  final String flow;
  late final HelperMethodVisitor _helperMethodVisitor;

  final List<MethodInvocation> allFlows = [];
  FunctionDeclaration? main;
  Map<String, List> get helperFunctions => _helperMethodVisitor.helperFunctions;
  final List<ImportDirective> imports = [];
  final List<FunctionDeclarationStatement> methodsDeclared = [];
  MethodInvocation? get getFlow {
    final index = allFlows.indexWhere((node) =>
        (node.argumentList.arguments.first as StringLiteral).stringValue ==
        flow);
    if (index != -1) {
      return allFlows[index];
    }
    return null;
  }

  @override
  visitMethodInvocation(MethodInvocation node) {
    if (node.methodName.name == 'testWidgets') {
      allFlows.add(node);
      if (node.flowDescription == flow) {
        node.visitChildren(_helperMethodVisitor);
      }
    }
  }

  @override
  visitImportDirective(ImportDirective node) {
    imports.add(node);
  }

  @override
  visitFunctionDeclarationStatement(FunctionDeclarationStatement node) {
    if (node.firstLayerToMain) {
      methodsDeclared.add(node);
    }
  }

  @override
  visitFunctionDeclaration(FunctionDeclaration node) {
    if (node.name.value() == 'main') {
      main = node;
    }
  }
}

class HelperMethodVisitor extends RecursiveAstVisitor {
  HelperMethodVisitor(this.sourceCode);
  final String sourceCode;
  Map<String, Map<String, String>> output = {};
  Map<String, List> helperFunctions = {};

  void handleBlock(FunctionExpression node, String type) {
    List<MapEntry<String, String>> testBlocks = [];

    String? currentComment;
    String currentCode = '';

    for (var statement in (node.body as BlockFunctionBody).block.statements) {
      final statementOffset = statement.offset;
      final statementEnd = statement.end;
      final codeLine = sourceCode.substring(statementOffset, statementEnd);
      final comment = statement.beginToken.precedingComments?.lexeme
          .replaceAll(RegExp(r'(/\*+|\*+/|^ *\*+ |^//)'), '')
          .trim();

      if (comment != null) {
        if (currentComment != null) {
          // Save the current block
          testBlocks.add(MapEntry(currentComment, currentCode));
        }

        // Start a new block
        currentComment = comment;
        currentCode = codeLine;
      } else {
        currentCode += '\n$codeLine';
      }
    }

    if (currentComment != null) {
      testBlocks.add(MapEntry(currentComment, currentCode));
    }

    output[type] = Map.fromEntries(testBlocks);
  }

  void handleBlockMethods(MethodInvocation node) {
    final element = node.methodName.staticElement;

    // get parent class
    ClassElement? parentClass;
    final enclosingElement = element!.enclosingElement;
    if (enclosingElement is ClassElement) {
      parentClass = enclosingElement;
    }

    final helperClassName = parentClass == null ? 'glboal' : parentClass.name;
    if (helperFunctions.containsKey(helperClassName)) {
      helperFunctions[helperClassName]!.add(element);
    } else {
      helperFunctions[helperClassName] = [element];
    }
  }

  @override
  visitMethodInvocation(MethodInvocation node) {
    final element = node.methodName.staticElement;
    final integration_test_path = PathUtils.integrationTestPath;
    if ((element?.librarySource?.fullName ?? '')
        .contains(integration_test_path)) {
      handleBlockMethods(node);
    }
    return super.visitMethodInvocation(node);
  }
}
